#pragma once

#include "Object.h"

namespace hyplex::drakan
{
	struct LevelMetadata : INode
	{
		struct DatabaseRef
		{
			unsigned int id = 0;
			String path;
		};

		String name;
		unsigned int maxWidth = 0;
		unsigned int maxHeight = 0;
		std::vector<DatabaseRef> dbRefs;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	struct LevelTerrain : INode
	{
		enum EType
		{
			FLOOR,
			CEILING,
			BETWEEN,
		};

		struct LayerVertex
		{
			static const unsigned int kFlagWater = 0x0001;
			static const unsigned int kFlagUnwalkable = 0x0100; // ?
			static const unsigned int kFlagUnknown0200 = 0x0200; // ? Either set for all or no vertices.
			uint16_t flags = 0;
			// height = (quantisedHeight + 32768) / 1024
			// height = (statusbar height - worldHeight) / 2048
			int16_t quantisedHeight = 0;
		};

		struct LayerGridCell
		{
			static const unsigned int kFlagDivisionIsDownRight = 1;
			static const unsigned int kFlagIsFlat = 2;

			// To see these in action, create a flat layer with a sharp cliff with a kink in it such that half a cell is walkable.
			static const unsigned int kFlagWalkable0 = 4;
			static const unsigned int kFlagWalkable1 = 8;

			uint16_t flags = 0;
			// 0 == 0.0f, 1 == 1.0f. 4 bits for 4 vertices.
			// Origin is bottom left (south west).
			// Vertices are top left, top right, bottom left, bottom right.
			uint8_t texCoordXBits = 0;
			uint8_t texCoordYBits = 0;
			uint32_t textureIds[2] = {};
		};

		struct Layer
		{
			uint32_t id = 0;
			EType type = FLOOR;
			String name;
			uivec2 size{};
			uivec2 position{};
			float worldHeight = 0;
			bool memberOfAlternateBlendingGroup = false; // ??
			float lightDirection = 0; // Probably
			float lightAscension = 0; // Probably
			std::array<uint8_t, 3> lightDiffuseColour{};
			std::array<uint8_t, 3> lightAmbientColour{};
			uint32_t unk0 = 0;
			std::vector<uint32_t> visibleLayerIndices;

			std::vector<LayerVertex> vertices;
			std::vector<LayerGridCell> cells;
		};

		std::vector<Layer> layers;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};

	struct LevelEntityCollection : INode
	{
		struct Entity
		{
			uint32_t id = 0;
			uint32_t objectId = 0;
			uint32_t lightingLayerId = 0;
			bool visible = false;
			uint16_t unk1C = 0;
			fvec3 position{};
			ivec3 rotationDegrees{};
			fvec3 scale{};
			std::vector<uint16_t> links;
			std::vector<Object::Property> props;
		};

		std::vector<Entity> entities;

		void consumeData(const Entry& entry, BasicBufferedLexState& bin);
		virtual ENodeType getType() const noexcept override;
		virtual void dumpNode(const NodeDumpArg& arg) const override;
	};
}
